﻿//-----------------------------------------------------------------------
// <copyright company="CoApp Project">
//     Copyright (c) 2010 Garrett Serack. All rights reserved.
//     Copyright (c) 2010 Eric Schultz. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

namespace CoApp.mkPackage {
    using System;
    using Toolkit.Extensions;
    using System.Collections.Generic;
    using System.Xml;
    using System.IO;
    using System.Xml.Xsl;
    using CoApp.Toolkit.Utility;
    using System.Xml.Serialization;
    using CoApp.Toolkit.Package;
    using System.Text;

    internal class Program {
      
   

        private bool _ignoreSigning = false;
        private bool _dontCreateMsi = false;
        private bool _outputIDizedSpec = false;
        private string _pfxPassword = null;
        private string _pfxFile = null;
        private string _outputFile = null;
        private bool _dontSignMsi = false;
        private bool _showTools = false;
        private string _outputFilePre = null;
        
        private ProcessUtility _candle = null;
        private ProcessUtility _light = null;

        
        private const string help = @"

Usage:
-------

mkPackage [options] <input-file>

    <input-file>                the file, conforming the mkPackage input schema, that describes the package
        
    Options:
    --------
    --help                      this help
    --nologo                    don't display the logo
    --load-config=<file>        loads configuration from <file>
    --verbose                   prints verbose messages
    
    --output-file=<file>        the destination package <file>

    --pfx-file=<file>           the path to a PFX file containing the signing certificate and key
    --pfx-password=<password>   the password to use the PFX file for signing

    --show-tools                displays the paths to the exe files used by mkPackage

    ----- only for debugging purposes -----
    --ignore-signing            Skip checking if binaries and assemblies are properly signed. 
                                Also skips signing the package
    --dont-create-msi           Don't create an package at the end
    --dont-sign-msi             Don't sign the package at the end
    --output-idized-spec        Outputs a modified version of the input file with the various 
                                generated IDs added

";
        

        private static int Main(string[] args) {
            return new Program().main(args);
        }

        private int main(string[] args) {
            
            var options = args.Switches();
            var parameters = new List<string>(args.Parameters());

            #region Parse Options 

            foreach(var arg in options.Keys) {
               
                var argumentParameters = (IList<string>)options[arg];

                switch(arg) {
                        /* options  */
                        
                    case "output-file":
                        this._outputFile = argumentParameters[0];
                        break;
                    
                        /* global switches */
                    case "load-config":
                        // all ready done, but don't get too picky.
                        break;

                    case "nologo":
                        this.Assembly().SetLogo("");
                        break;

                    case "help":
                        return Help();
                        
                    case "ignore-signing":
                        this._ignoreSigning = true;
                        break;

                    case "dont-create-msi":
                        this._dontCreateMsi = true;
                        break;

                    case "output-idized-spec":
                        this._outputIDizedSpec = true;
                        break;
                    case "dont-sign-msi":
                        this._dontSignMsi = true;
                        break;
                    
                    case "pfx-file":
                        this._pfxFile = argumentParameters[0];
                        break;
                   
                    case "pfx-password":
                        this._pfxPassword = argumentParameters[0];
                        break;

                    case "show-tools":
                        this._showTools = true;
                        break;

                    default:
                        return Fail("Unknown parameter [--{0}]", arg);
                }
            }
            Logo();

            #endregion

            /*
            if(parameters.Count != 1) {

                return Fail("Missign source code root path. \r\n\r\n    Use --help for command line help.");
            }*/

            Console.WriteLine("Looking up tool locations...");

            _candle = new ProcessUtility(ProgramFinder.ProgramFilesAndDotNet.ScanForFile("candle.exe"));
            _light = new ProcessUtility(ProgramFinder.ProgramFilesAndDotNet.ScanForFile("light.exe"));

            if (_candle.Executable == null)
                return Fail("candle.exe from WiX could not be found. Are you sure it's installed?");

            if (_light.Executable == null)
                return Fail("light.exe from WiX could not be found. Are you sure it's installed?");

            if (_showTools)
            {
                Console.WriteLine(_candle.Executable);
                Console.WriteLine(_light.Executable);
            }


            Console.WriteLine("Running mkPackage on: {0}", parameters[0]);
            //XmlDocument specFile = null;
            Package p = null;
            var xs = new XmlSerializer(typeof(Package));
            try
            {
                using (var r = new StreamReader(parameters[0]))
                {
                    p = (Package)xs.Deserialize(r);
                }
            }
            catch (IOException)
            {
                return Fail("Package specification file couldn't be opened.");
            }
            catch (InvalidOperationException e)
            {
                if (e.InnerException.GetType() == typeof(XmlException))
                {
                    return Fail("There was a problem with the syntax of your file. The error is as follows: {0}",
                                e.InnerException.Message);
                }

                return Fail("There was an unspecified problem with the file.");
            }
            catch (Exception e)
            {
                return Fail("There was an unspecified problem with the file.");
            }
            
     



            var idizer = new IDizer(p);
            if (_outputFile == null)
            {
                _outputFilePre = String.Format("{0}-{1}-{2}", p.Name,p.Version, p.Arch);
                _outputFile = _outputFilePre + ".msi";
            }

            

           
           
            if (!idizer.addIDs())
                return Fail("There was a problem with your XML syntax.");

            //Verify that all binary files have signatures and assms are signed by the inputted PFX
            if (!this._ignoreSigning)
            {
                //TODO: Verifying x64 binaries does not work... also we don't/can't verify .cat files
                var pfxVerifier = new PKCSVerifier(this._pfxFile, this._pfxPassword);
                if (p.Arch != Arch.x64 && (!pfxVerifier.VerifyAssemblies(idizer.CollectAssmFiles()) ||
                    !pfxVerifier.VerifyNonAssemblies(idizer.CollectBinFiles())))
                    return Fail("File signatures could not be verified. Make sure your binary files (.EXE, .DLL, .OCX, etc.) are digitally signed.");
            }


            //convert package back to xml!
            var specFile = new XmlDocument();
            
            StringBuilder memoryStream = null;
            try
            {

                memoryStream = new StringBuilder();
                var xw = XmlWriter.Create(memoryStream);
                
                
                xs.Serialize(xw, p);
                xw.Flush();

                specFile.LoadXml(memoryStream.ToString());
                
                if (_outputIDizedSpec)
                {
                    using (var streamwriter = new StreamWriter(_outputFilePre + "spec.xml", false))
                    {
                        streamwriter.Write(memoryStream);
                    }
                }

            }
            catch (Exception e)
            {
                return Fail("Internal failure converting back to XML " + 
                    Environment.NewLine + 
                    "Exception type: {0}; Message:{1}", e.GetType().Name, e.Message);

            }
            
           
            //author wix
            var tempPrefix = Path.GetTempFileName();
            var xslArgs = new XsltArgumentList();
            xslArgs.AddParam("bootstrap-src", "", tempPrefix + "bootstrap.exe");
            using (var fs = System.IO.File.Create(tempPrefix + "bootstrap.exe"))
            {
                fs.Write(Properties.Resources.coapp_bootstrap, 0, Properties.Resources.coapp_bootstrap.Length);
            }
            
            var transform = new XslCompiledTransform(true);
            using (TextReader tr = new StringReader(CoApp.mkPackage.Properties.Resources.PkgToWix))
            {
                using (var r = XmlReader.Create(tr))
                {
                    transform.Load(r);
                }
            }

            tempPrefix = Path.GetTempFileName();
            using (var fs = System.IO.File.Create(tempPrefix + ".wxs", 8192))
            {
                
               
                transform.Transform(specFile, xslArgs, fs);
            }
            //create MSI

            if (!_dontCreateMsi)
            {

                
                //we suppress the lack of UpgradeCode warning since we don't use them
                int ret = _candle.Exec("-nologo -sw1075 -out {0}.wixobj {0}.wxs", tempPrefix);

                if (ret != 0)
                    return Fail("Candle failed. Something screwy is going on." + Environment.NewLine + _candle.StandardOut);

                
                //do the compile
              
                
                //we suppress the lack of UpgradeCode warning since we don't use them
                ret = _light.Exec("-nologo -sw1076 -out {1} {0}.wixobj", tempPrefix, _outputFile);

                if (ret != 0)
                    return Fail("Light failed. Something screwy is going on." + Environment.NewLine + _light.StandardOut);
            }

           
           
            //digitally sign the msi
            if (!_dontCreateMsi && (!_ignoreSigning || !_dontSignMsi ) )
            {
                var s = new PfxSigner(_pfxFile, _pfxPassword);
                if (!s.Sign(_outputFile))
                    return Fail("The MSI couldn't be signed.");
            }

            return 0;
        }





       

        #region fail/help/logo

        public int Fail(string text, params object[] par) {
            Logo();
            using(new ConsoleColors(ConsoleColor.Red, ConsoleColor.Black))
                Console.WriteLine("Error:{0}", text.format(par));
            return 1;
        }

        private int Help() {
            Logo();
            using(new ConsoleColors(ConsoleColor.White, ConsoleColor.Black))
                help.Print();
            return 0;
        }

        private void Logo() {
            using(new ConsoleColors(ConsoleColor.Cyan, ConsoleColor.Black))
                this.Assembly().Logo().Print();
            this.Assembly().SetLogo("");
        }

        #endregion


       

        /*
        /// <summary>
        /// Collects all the assemblies and returns them.
        /// </summary>
        /// <param name="doc">The Document</param>
        /// <param name="idizer"></param>
        /// <returns>A List of the file paths to all the assemblies in the input document</returns>
        private List<String> CollectAssmFiles(XmlDocument doc, IDizer idizer)
        {
            var nodes = idizer.SelectNodes(doc, xpathToAssm);
            var output = new List<String>();
            foreach (XmlNode node in nodes)
            {
                output.Add(node.Attributes[srcAttribOnBin].Value);
            }

            return output;
        }

        private List<String> CollectBinFiles(XmlDocument doc, IDizer idizer)
        {
            var nodes = idizer.SelectNodes(doc, xpathToAssm);
            var output = new List<String>();
            foreach (XmlNode node in nodes)
            {
                FileInfo fi = new FileInfo(node.Attributes[srcAttribOnBin].Value);

                if (binExts.Contains(fi.Extension.ToLower()))
                {
                    output.Add(fi.FullName);
                }
            }

            return output;
        }

        private XmlDocument LoadSpecFile(string specFilePath)
        {
            var specFile = new XmlDocument();
            specFile.Load(specFilePath);
            return specFile;
        }


        private bool VerifyFileConformance(XmlDocument doc)
        {
            try
            {
                if (doc.Schemas == null)
                    doc.Schemas = new XmlSchemaSet();
                

                using(TextReader r = new StringReader(CoApp.Toolkit.Properties.Resources.mkPackage))
                {
                    XmlSchema schema = XmlSchema.Read(r, null);
                    doc.Schemas.Add(schema);
                }

                doc.Validate(null);
            }
            catch
            {
                return false;
            }
            return true;
        }



        private void addFeature(XmlDocument doc, List<string> compIds)
        {
            XmlNamespaceManager mgr = new XmlNamespaceManager(doc.NameTable);
            mgr.AddNamespace("in", schemaUri);
            var packageElem = (XmlElement)doc.SelectSingleNode("//in:Package", mgr);
            var featuresElem = doc.CreateElement("features", schemaUri);

            packageElem.AppendChild(featuresElem);

            foreach (string s in compIds)
            {
                var temp = doc.CreateElement("comp", schemaUri);
                temp.SetAttribute("component_id", s);
                featuresElem.AppendChild(temp);
            }
        }*/
    }
}